"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const analysis_1 = require("./analysis");
const config_1 = require("./config");
const utils_1 = require("./utils");
const diagnosticMessageCodes = [100000, 100001, 100002, 100003];
const diagnosticMessages = {
    100000: {
        messageText: (e) => `Failed to parse: ${e.message}.`,
        category: "Error",
    },
    100001: {
        messageText: (e) => `Failed to parse: ${e.message}.`,
        category: "Error",
    },
    100002: {
        messageText: (p) => `Cannot find type for parameter ${stringifyParameter(p)} in schema.`,
        category: "Error",
    },
    100003: {
        messageText: (originalMessage) => `There was an issue type checking this expression. Original error: ${originalMessage}`,
        category: "Warning",
    },
};
const unsupportedTypeScriptErrors = new Set([
    // "Cannot find name '{0}'."
    // The plugin tries to resolve all type names and create a literal type.
    // This type is then checked against the type from the DB schema. If we get
    // this error, it most likely means not all type names were resolved
    // correctly.
    2304,
]);
const getTemplateExpressions = (typescript, node) => typescript.isTemplateExpression(node)
    ? node.templateSpans.map((span) => span.expression)
    : [];
const stringifyParameter = (parameter) => [
    parameter.schema,
    parameter.table,
    parameter.column,
    parameter.jsonPath && parameter.jsonPath.path,
]
    .filter((x) => x)
    .join(".");
const getParameterType = (parameter, schemaJson, defaultSchemaName) => {
    const schema = parameter.schema || defaultSchemaName;
    const dbSchema = schemaJson[schema];
    const dbTable = dbSchema && dbSchema[parameter.table];
    const dbColumn = dbTable && dbTable[parameter.column];
    if (dbColumn) {
        const type = parameter.jsonPath && parameter.jsonPath.isText
            ? "string | null"
            : dbColumn;
        return parameter.isArray ? `Array<${type}>` : type;
    }
};
const getDiagnosticFactory = (context) => {
    const file = context.node.getSourceFile();
    const source = config_1.pluginName;
    const any = ({ code, messageText, category, start, length, }) => ({
        code,
        messageText,
        category,
        start,
        length,
        file,
        source,
    });
    const own = (code, arg, pos) => any({
        code,
        messageText: diagnosticMessages[code].messageText(arg),
        category: context.typescript.DiagnosticCategory[diagnosticMessages[code].category],
        ...pos,
    });
    const pos = (expression) => ({
        start: expression.getStart(file) - context.node.getStart(file) - 1,
        length: expression.getWidth(file),
    });
    return {
        any,
        own,
        pos,
    };
};
class SqlTemplateLanguageService {
    constructor(logger, config, typeChecker, typeResolver) {
        this.logger = logger;
        this.config = config;
        this.typeChecker = typeChecker;
        this.typeResolver = typeResolver;
    }
    getSemanticDiagnostics(context) {
        const factory = getDiagnosticFactory(context);
        let analysis;
        try {
            analysis = analysis_1.analyze(context.text);
        }
        catch (e) {
            return e instanceof analysis_1.ParseError
                ? [factory.own(100001, e, { start: e.cursorPosition - 1, length: 1 })]
                : [factory.own(100000, e, { start: 0, length: context.text.length })];
        }
        for (const warning of analysis.warnings) {
            this.logger.log(`Warning analyzing template in file ${context.fileName}: ${warning}`);
        }
        const schema = this.config.schema;
        if (!schema) {
            return [];
        }
        const expressions = getTemplateExpressions(context.typescript, context.node);
        const diagnostics = Array.from(analysis.parameters.entries())
            .map(([index, parameter]) => ({
            expression: expressions[index - 1],
            parameter,
        }))
            .map(({ expression, parameter }) => {
            const parameterType = getParameterType(parameter, schema, this.config.defaultSchemaName);
            if (!parameterType) {
                return [factory.own(100002, parameter, factory.pos(expression))];
            }
            const expressionType = this.typeResolver.getType(expression);
            const content = `{ let expr: ${expressionType} = null as any; let param: ${parameterType} = expr; }`;
            return this.typeChecker.check(content).map((diagnostic) => unsupportedTypeScriptErrors.has(diagnostic.code)
                ? factory.own(100003, diagnostic.messageText, factory.pos(expression))
                : factory.any({
                    code: diagnostic.code,
                    messageText: diagnostic.messageText,
                    category: diagnostic.category,
                    ...factory.pos(expression),
                }));
        });
        return utils_1.flatten(diagnostics);
    }
}
exports.default = SqlTemplateLanguageService;
//# sourceMappingURL=language-service.js.map